@@ -26,6 +26,7 @@ import logging
 import re
 import urllib
 import xml.etree.ElementTree
+from functools import lru_cache
 from html.parser import HTMLParser
 from urllib.parse import parse_qs
 
@@ -147,8 +148,21 @@ hls_formats = [
 ]
 hls_formats_dict = dict(hls_formats)
 
-V3_API_ENDPOINT = 'https://www.googleapis.com/youtube/v3'
 CHANNEL_VIDEOS_XML = 'https://www.youtube.com/feeds/videos.xml'
+WATCH_ENDPOINT = 'https://www.youtube.com/watch?bpctr=9999999999&has_verified=1&v='
+
+# The page may contain "};" sequences inside the initial player response.
+# Use a greedy match with script end tag, and fallback to a non-greedy match without.
+INITIAL_PLAYER_RESPONSE_RE1 = r'ytInitialPlayerResponse\s*=\s*({.+})\s*;\s*</script'
+INITIAL_PLAYER_RESPONSE_RE2 = r'ytInitialPlayerResponse\s*=\s*({.+?})\s*;'
+
+
+def get_ipr(page):
+    for regex in (INITIAL_PLAYER_RESPONSE_RE1, INITIAL_PLAYER_RESPONSE_RE2):
+        ipr = re.search(regex, page)
+        if ipr is not None:
+            return ipr
+    return None
 
 
 class YouTubeError(Exception):
@@ -198,12 +212,12 @@ def youtube_get_old_endpoint(vid):
 
 
 def youtube_get_new_endpoint(vid):
-    url = 'https://www.youtube.com/watch?bpctr=9999999999&has_verified=1&v=' + vid
+    url = WATCH_ENDPOINT + vid
     r = util.urlopen(url)
     if not r.ok:
         raise YouTubeError('Youtube "%s": %d %s' % (url, r.status_code, r.reason))
 
-    ipr = re.search(r'ytInitialPlayerResponse\s*=\s*({.+?})\s*;', r.text)
+    ipr = get_ipr(r.text)
     if ipr is None:
         try:
             url = get_gdpr_consent_url(r.text)
@@ -213,13 +227,41 @@ def youtube_get_new_endpoint(vid):
         if not r.ok:
             raise YouTubeError('Youtube "%s": %d %s' % (url, r.status_code, r.reason))
 
-        ipr = re.search(r'ytInitialPlayerResponse\s*=\s*({.+?})\s*;', r.text)
+        ipr = get_ipr(r.text)
         if ipr is None:
             raise YouTubeError('Youtube "%s": No ytInitialPlayerResponse found' % url)
 
     return None, ipr.group(1)
 
 
+def get_total_time(episode):
+    try:
+        vid = get_youtube_id(episode.url)
+        if vid is None:
+            return 0
+
+        url = WATCH_ENDPOINT + vid
+        r = util.urlopen(url)
+        if not r.ok:
+            return 0
+
+        ipr = get_ipr(r.text)
+        if ipr is None:
+            url = get_gdpr_consent_url(r.text)
+            r = util.urlopen(url)
+            if not r.ok:
+                return 0
+
+            ipr = get_ipr(r.text)
+            if ipr is None:
+                return 0
+
+        player_response = json.loads(ipr.group(1))
+        return int(player_response['videoDetails']['lengthSeconds'])  # 0 if live
+    except:
+        return 0
+
+
 def get_real_download_url(url, allow_partial, preferred_fmt_ids=None):
     if not preferred_fmt_ids:
         preferred_fmt_ids, _, _ = formats_dict[22]  # MP4 720p
@@ -325,11 +367,8 @@ def get_real_download_url(url, allow_partial, preferred_fmt_ids=None):
     return url, duration
 
 
+@lru_cache(1)
 def get_youtube_id(url):
-    r = re.compile(r'http[s]?://(?:[a-z]+\.)?youtube\.com/v/(.*)\.swf', re.IGNORECASE).match(url)
-    if r is not None:
-        return r.group(1)
-
     r = re.compile(r'http[s]?://(?:[a-z]+\.)?youtube\.com/watch\?v=([^&]*)', re.IGNORECASE).match(url)
     if r is not None:
         return r.group(1)
@@ -338,6 +377,10 @@ def get_youtube_id(url):
     if r is not None:
         return r.group(1)
 
+    r = re.compile(r'http[s]?://(?:[a-z]+\.)?youtube\.com/v/(.*)\.swf', re.IGNORECASE).match(url)
+    if r is not None:
+        return r.group(1)
+
     return for_each_feed_pattern(lambda url, channel: channel, url, None)
 
 
@@ -361,6 +404,7 @@ def for_each_feed_pattern(func, url, fallback_result):
         r'http[s]?://(?:[a-z]+\.)?youtube\.com/profile?user=([a-z0-9]+)',
         r'http[s]?://(?:[a-z]+\.)?youtube\.com/rss/user/([a-z0-9]+)/videos\.rss',
         r'http[s]?://(?:[a-z]+\.)?youtube\.com/channel/([-_a-z0-9]+)',
+        r'http[s]?://(?:[a-z]+\.)?youtube\.com/feeds/videos.xml\?user=([a-z0-9]+)',
         r'http[s]?://(?:[a-z]+\.)?youtube\.com/feeds/videos.xml\?channel_id=([-_a-z0-9]+)',
         r'http[s]?://gdata.youtube.com/feeds/users/([^/]+)/uploads',
         r'http[s]?://gdata.youtube.com/feeds/base/users/([^/]+)/uploads',
@@ -385,26 +429,41 @@ def get_real_channel_url(url):
     return for_each_feed_pattern(return_user_feed, url, url)
 
 
-def get_channel_id_url(url):
+@lru_cache(1)
+def get_channel_id_url(url, feed_data=None):
     if 'youtube.com' in url:
         try:
-            req = util.urlopen(url)
+            if feed_data is None:
+                r = util.urlopen(url)
+                if not r.ok:
+                    raise YouTubeError('Youtube "%s": %d %s' % (url, r.status_code, r.reason))
+            else:
+                r = feed_data
             # video page may contain corrupt HTML/XML, search for tag to avoid exception
-            m = re.search(r'<meta itemprop="channelId" content="([^"]+)">', req.text)
+            m = re.search(r'channel_id=([^"]+)">', r.text)
             if m:
                 channel_id = m.group(1)
             else:
-                raw_xml_data = io.BytesIO(req.content)
+                raw_xml_data = io.BytesIO(r.content)
                 xml_data = xml.etree.ElementTree.parse(raw_xml_data)
                 channel_id = xml_data.find("{http://www.youtube.com/xml/schemas/2015}channelId").text
+                if channel_id is None:
+                    # check entries if feed has an empty channelId
+                    m = re.search(r'<yt:channelId>([^<]+)</yt:channelId>', r.text)
+                    if m:
+                        channel_id = m.group(1)
+                    if channel_id is None:
+                        raise Exception('Could not retrieve YouTube channel ID for URL %s.' % url)
             channel_url = 'https://www.youtube.com/channel/{}'.format(channel_id)
             return channel_url
 
         except Exception:
-            logger.warning('Could not retrieve youtube channel id.', exc_info=True)
+            logger.warning('Could not retrieve YouTube channel ID for URL %s.' % url, exc_info=True)
+
+    raise Exception('Could not retrieve YouTube channel ID for URL %s.' % url)
 
 
-def get_cover(url):
+def get_cover(url, feed_data=None):
     if 'youtube.com' in url:
 
         class YouTubeHTMLCoverParser(HTMLParser):
@@ -429,8 +488,11 @@ def get_cover(url):
                     self.url.append(attribute_dict['src'])
 
         try:
-            channel_url = get_channel_id_url(url)
-            html_data = util.response_text(util.urlopen(channel_url))
+            channel_url = get_channel_id_url(url, feed_data)
+            r = util.urlopen(channel_url)
+            if not r.ok:
+                raise YouTubeError('Youtube "%s": %d %s' % (url, r.status_code, r.reason))
+            html_data = util.response_text(r)
             parser = YouTubeHTMLCoverParser()
             parser.feed(html_data)
             if parser.url:
@@ -477,11 +539,11 @@ def get_gdpr_consent_url(html_data):
         logger.debug('YouTube GDPR accept consent URL is: %s', parser.url)
         return parser.url
     else:
-        logger.debug('YouTube GDPR accepted consent URL could not be resolved.', parser.url)
+        logger.debug('YouTube GDPR accepted consent URL could not be resolved.')
         raise YouTubeError('No acceptable GDPR consent URL')
 
 
-def get_channel_desc(url):
+def get_channel_desc(url, feed_data=None):
     if 'youtube.com' in url:
 
         class YouTubeHTMLDesc(HTMLParser):
@@ -500,8 +562,11 @@ def get_channel_desc(url):
                     self.description = attribute_dict['content']
 
         try:
-            channel_url = get_channel_id_url(url)
-            html_data = util.response_text(util.urlopen(channel_url))
+            channel_url = get_channel_id_url(url, feed_data)
+            r = util.urlopen(channel_url)
+            if not r.ok:
+                raise YouTubeError('Youtube "%s": %d %s' % (url, r.status_code, r.reason))
+            html_data = util.response_text(r)
             parser = YouTubeHTMLDesc()
             parser.feed(html_data)
             if parser.description:
